;; Generated package description from cape.el  -*- no-byte-compile: t -*-
(define-package "cape" "0.17.0.20231127.191212" "Completion At Point Extensions" '((emacs "27.1") (compat "29.1.4.2")) :commit "d63052ac2a3d8193d5df60568b74b47c4d4fed50" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("abbrev" "convenience" "matching" "completion" "wp") :url "https://github.com/minad/cape")
